import logging
import redis
import paho.mqtt.client as mqtt

# Custom logging handler for MQTT.
class MQTTLoggingHandler(logging.Handler):
    def __init__(self, broker='localhost', port=1883, topic='logs'):
        super().__init__()
        self.topic = topic
        self.client = mqtt.Client()
        # Connect to the MQTT broker.
        self.client.connect(broker, port)
        # Start a background network loop.
        self.client.loop_start()

    def emit(self, record):
        try:
            log_entry = self.format(record)
            self.client.publish(self.topic, log_entry)
        except Exception:
            self.handleError(record)

# Custom logging handler for Redis.
class RedisLoggingHandler(logging.Handler):
    def __init__(self, host='localhost', port=6379, channel='logs'):
        super().__init__()
        self.redis_client = redis.Redis(host=host, port=port, decode_responses=True)
        self.channel = channel

    def emit(self, record):
        try:
            log_entry = self.format(record)
            self.redis_client.publish(self.channel, log_entry)
        except Exception:
            self.handleError(record)

# Logging and Tracking Handler that wraps one active handler based on configuration.
class LoggingAndTrackingHandler(logging.Handler):
    def __init__(self, config):
        """
        Initialize the handler based on the provided configuration.

        :param config: Dictionary with logging configuration.
            Example:
            {
                'log_level': 'DEBUG',
                'logger_type': 'file',  # Options: 'terminal', 'file', 'mqtt', 'redis'
                'terminal': {},
                'file': {'filename': 'app.log'},
                'mqtt': {'broker': 'localhost', 'port': 1883, 'topic': 'logs'},
                'redis': {'host': 'localhost', 'port': 6379, 'channel': 'logs'}
            }
        """
        super().__init__()
        self.config = config
        self.level = getattr(logging, config.get('log_level', 'INFO').upper(), logging.INFO)
        self.active_handler = self._setup_handler()

    def _setup_handler(self):
        active_type = self.config.get('logger_type', 'terminal').lower()
        if active_type == 'terminal':
            handler = logging.StreamHandler()
        elif active_type == 'file':
            file_config = self.config.get('file', {})
            if 'filename' not in file_config:
                raise ValueError("File handler selected but no filename provided in config.")
            handler = logging.FileHandler(file_config['filename'])
        elif active_type == 'mqtt':
            mqtt_config = self.config.get('mqtt', {})
            handler = MQTTLoggingHandler(
                broker=mqtt_config.get('broker', 'localhost'),
                port=mqtt_config.get('port', 1883),
                topic=mqtt_config.get('topic', 'logs')
            )
        elif active_type == 'redis':
            redis_config = self.config.get('redis', {})
            handler = RedisLoggingHandler(
                host=redis_config.get('host', 'localhost'),
                port=redis_config.get('port', 6379),
                channel=redis_config.get('channel', 'logs')
            )
        else:
            raise ValueError(f"Unknown logger_type specified: {active_type}")
        return handler

    def setFormatter(self, fmt):
        """Override setFormatter to also set the formatter on the active handler."""
        super().setFormatter(fmt)
        self.active_handler.setFormatter(fmt)

    def emit(self, record):
        """Delegate emit to the wrapped active handler."""
        self.active_handler.emit(record)

# Example usage:
if __name__ == '__main__':
    config = {
        'log_level': 'DEBUG',
        'logger_type': 'redis',  # Change to 'terminal', 'file', 'mqtt', or 'redis' as needed.
        'terminal': {},
        'file': {'filename': 'app.log'},
        'mqtt': {'broker': 'localhost', 'port': 1883, 'topic': 'logs'},
        'redis': {'host': 'localhost', 'port': 6379, 'channel': 'logs'}
    }

    # Create the main logger.
    logger = logging.getLogger("main")
    logger.setLevel(logging.DEBUG)

    # Instantiate the custom logging handler.
    custom_handler = LoggingAndTrackingHandler(config)

    # Set the formatter here in main (instead of inside the class).
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    formatter = logging.Formatter(log_format)
    custom_handler.setFormatter(formatter)

    logger.addHandler(custom_handler)

    # Log some test messages.
    logger.debug("This is a debug message.")
    logger.info("This is an info message.")
    logger.warning("This is a warning message.")
    logger.error("This is an error message.")
    logger.critical("This is a critical message.")



import time
import string
import random

# Constants for configuration keys
KEY_KNOWLEDGE_TYPE = "knowledge_type"
KEY_HOST = "host"
KEY_PORT = "port"
KEY_DB = "db"
KEY_TOPIC = "topic"
KEY_BOOTSTRAP_SERVERS = "bootstrap_servers"

# Constants for backends
BACKEND_REDIS = "redis"
BACKEND_MEMCACHED = "memcached"
BACKEND_IGNITE = "ignite"
BACKEND_HAZELCAST = "hazelcast"
BACKEND_TARANTOOL = "tarantool"
BACKEND_AEROSPIKE = "aerospike"
BACKEND_KAFKA = "kafka"
BACKEND_SQLITE = "sqlite"

# Constants for SQLite queries and in-memory DB
SQLITE_MEMORY = ":memory:"
SQLITE_CREATE_TABLE = "CREATE TABLE kv (key TEXT PRIMARY KEY, value TEXT)"
SQLITE_INSERT_OR_REPLACE = "INSERT OR REPLACE INTO kv (key, value) VALUES (?, ?)"
SQLITE_SELECT_VALUE = "SELECT value FROM kv WHERE key=?"
SQLITE_DELETE_KEY = "DELETE FROM kv WHERE key=?"

# Other constants
ERROR_UNSUPPORTED_BACKEND = "Unsupported backend: {}"
DEFAULT_CACHE_NAME = "default"
AEROSPIKE_VALUE_FIELD = "value"

# Helper function to generate random strings (if needed)
def random_string(length=10):
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))


class KnowledgeManager:
    def __init__(self, config):
        """
        Initialize the KnowledgeManager with a chosen backend.
        Supported backends: redis, memcached, ignite, hazelcast, tarantool, aerospike, kafka, sqlite

        :param config: A dictionary containing all configuration parameters needed by the backend.
        """
        self.backend = config[KEY_KNOWLEDGE_TYPE]
        if self.backend == BACKEND_REDIS:
            import redis
            self.client = redis.Redis(host=config[KEY_HOST], port=config[KEY_PORT], db=config[KEY_DB])
        elif self.backend == BACKEND_MEMCACHED:
            import memcache
            self.client = memcache.Client([(config[KEY_HOST], config[KEY_PORT])])
        elif self.backend == BACKEND_IGNITE:
            from pyignite import Client
            self.client = Client()
            self.client.connect(config[KEY_HOST], config[KEY_PORT])
        elif self.backend == BACKEND_HAZELCAST:
            import hazelcast
            # Configuration for hazelcast client is passed via config.
            self.client = hazelcast.HazelcastClient(**config)
        elif self.backend == BACKEND_TARANTOOL:
            import tarantool
            self.client = tarantool.Connection(config[KEY_HOST], config[KEY_PORT])
        elif self.backend == BACKEND_AEROSPIKE:
            import aerospike
            cfg = {'hosts': [(config[KEY_HOST], config[KEY_PORT])]}
            self.client = aerospike.client(cfg).connect()
        elif self.backend == BACKEND_KAFKA:
            from kafka import KafkaProducer, KafkaConsumer
            self.topic = config[KEY_TOPIC]
            self.producer = KafkaProducer(bootstrap_servers=config[KEY_BOOTSTRAP_SERVERS])
            self.consumer = KafkaConsumer(
                self.topic,
                bootstrap_servers=config[KEY_BOOTSTRAP_SERVERS],
                auto_offset_reset='earliest',
                consumer_timeout_ms=1000
            )
        elif self.backend == BACKEND_SQLITE:
            import sqlite3
            self.conn = sqlite3.connect(SQLITE_MEMORY)
            self.cursor = self.conn.cursor()
            self.cursor.execute(SQLITE_CREATE_TABLE)
        else:
            raise ValueError(ERROR_UNSUPPORTED_BACKEND.format(self.backend))

    def write(self, key, value):
        """Write a key-value pair to the backend."""
        if self.backend == BACKEND_REDIS:
            return self.client.set(key, value)
        elif self.backend == BACKEND_MEMCACHED:
            return self.client.set(key, value)
        elif self.backend == BACKEND_IGNITE:
            cache = self.client.get_or_create_cache(DEFAULT_CACHE_NAME)
            return cache.put(key, value)
        elif self.backend == BACKEND_HAZELCAST:
            hz_map = self.client.get_map(DEFAULT_CACHE_NAME).blocking()
            return hz_map.put(key, value)
        elif self.backend == BACKEND_TARANTOOL:
            # Assumes a space 'kv' exists with schema (key, value)
            return self.client.insert('kv', (key, value))
        elif self.backend == BACKEND_AEROSPIKE:
            # Note: In Aerospike, the key should be a tuple; this is a simplified example.
            return self.client.put(key, {AEROSPIKE_VALUE_FIELD: value})
        elif self.backend == BACKEND_KAFKA:
            # Kafka is a streaming system; here, writing means sending a message.
            future = self.producer.send(self.topic, key=key.encode(), value=value.encode())
            return future.get(timeout=10)
        elif self.backend == BACKEND_SQLITE:
            try:
                self.cursor.execute(SQLITE_INSERT_OR_REPLACE, (key, value))
                self.conn.commit()
                return True
            except Exception as e:
                self.conn.rollback()
                return False

    def read(self, key, queueSize=1):
        """Read the value for a given key from the backend."""
        if self.backend == BACKEND_REDIS:
            return self.client.get(key)
        elif self.backend == BACKEND_MEMCACHED:
            return self.client.get(key)
        elif self.backend == BACKEND_IGNITE:
            cache = self.client.get_or_create_cache(DEFAULT_CACHE_NAME)
            return cache.get(key)
        elif self.backend == BACKEND_HAZELCAST:
            hz_map = self.client.get_map(DEFAULT_CACHE_NAME).blocking()
            return hz_map.get(key)
        elif self.backend == BACKEND_TARANTOOL:
            result = self.client.select('kv', 0, 1, 0, key)
            return result[0][1] if result else None
        elif self.backend == BACKEND_AEROSPIKE:
            try:
                (key_rec, metadata, record) = self.client.get(key)
                return record.get(AEROSPIKE_VALUE_FIELD)
            except Exception as e:
                return None
        elif self.backend == BACKEND_KAFKA:
            # For Kafka, simulate reading by consuming one message from the topic.
            for message in self.consumer:
                return message.value.decode()
            return None
        elif self.backend == BACKEND_SQLITE:
            self.cursor.execute(SQLITE_SELECT_VALUE, (key,))
            row = self.cursor.fetchone()
            return row[0] if row else None

    def delete(self, key):
        """Delete the key-value pair identified by key from the backend."""
        if self.backend == BACKEND_REDIS:
            return self.client.delete(key)
        elif self.backend == BACKEND_MEMCACHED:
            return self.client.delete(key)
        elif self.backend == BACKEND_IGNITE:
            cache = self.client.get_or_create_cache(DEFAULT_CACHE_NAME)
            return cache.remove(key)
        elif self.backend == BACKEND_HAZELCAST:
            hz_map = self.client.get_map(DEFAULT_CACHE_NAME).blocking()
            return hz_map.remove(key)
        elif self.backend == BACKEND_TARANTOOL:
            return self.client.delete('kv', key)
        elif self.backend == BACKEND_AEROSPIKE:
            try:
                return self.client.remove(key)
            except Exception as e:
                return None
        elif self.backend == BACKEND_KAFKA:
            # Kafka does not support deleting individual messages.
            return None
        elif self.backend == BACKEND_SQLITE:
            self.cursor.execute(SQLITE_DELETE_KEY, (key,))
            self.conn.commit()
            return True

    def exists(self, key):
        """Check whether a key exists in the backend."""
        if self.backend == BACKEND_REDIS:
            return self.client.exists(key)
        elif self.backend == BACKEND_MEMCACHED:
            return self.client.get(key) is not None
        elif self.backend == BACKEND_IGNITE:
            cache = self.client.get_or_create_cache(DEFAULT_CACHE_NAME)
            return cache.get(key) is not None
        elif self.backend == BACKEND_HAZELCAST:
            hz_map = self.client.get_map(DEFAULT_CACHE_NAME).blocking()
            return hz_map.contains_key(key)
        elif self.backend == BACKEND_TARANTOOL:
            result = self.client.select('kv', 0, 1, 0, key)
            return bool(result)
        elif self.backend == BACKEND_AEROSPIKE:
            try:
                self.client.get(key)
                return True
            except Exception as e:
                return False
        elif self.backend == BACKEND_KAFKA:
            # Kafka does not provide a means to check for existence of a message by key.
            return None
        elif self.backend == BACKEND_SQLITE:
            self.cursor.execute('SELECT 1 FROM kv WHERE key=?', (key,))
            return self.cursor.fetchone() is not None


def benchmark_operations(backend, config, num_operations=1000):
    print(f"\nBenchmarking backend: {backend.upper()} with {num_operations} operations")
    try:
        km = KnowledgeManager(config)
    except Exception as e:
        print(f"Could not initialize backend '{backend}': {e}")
        return

    keys = [f'key_{i}' for i in range(num_operations)]
    value = "sample_value"

    # Write benchmark
    start_time = time.time()
    for key in keys:
        km.write(key, value)
    write_time = time.time() - start_time
    print(f"Write: {write_time:.6f} sec total, {(write_time/num_operations)*1e6:.2f} µs/op")

    # Read benchmark
    start_time = time.time()
    for key in keys:
        km.read(key)
    read_time = time.time() - start_time
    print(f"Read: {read_time:.6f} sec total, {(read_time/num_operations)*1e6:.2f} µs/op")

    # Exists benchmark
    start_time = time.time()
    for key in keys:
        km.exists(key)
    exists_time = time.time() - start_time
    print(f"Exists: {exists_time:.6f} sec total, {(exists_time/num_operations)*1e6:.2f} µs/op")

    # Delete benchmark
    start_time = time.time()
    for key in keys:
        km.delete(key)
    delete_time = time.time() - start_time
    print(f"Delete: {delete_time:.6f} sec total, {(delete_time/num_operations)*1e6:.2f} µs/op")


if __name__ == '__main__':
    # Define backend configurations here. Adjust values as needed for your environment.
    backend_configs = {
        'redis': {"host": "localhost", "port": 6379, "db": 0},
        # 'memcached': {"host": "127.0.0.1", "port": 11211},
        # 'ignite': {"host": "127.0.0.1", "port": 10800}, #TODO:Test this
        # 'hazelcast': {},  # Provide Hazelcast-specific configuration if needed. #TODO:Test this
        # 'tarantool': {"host": "127.0.0.1", "port": 3301}, #TODO:Test this
        # 'aerospike': {"host": "127.0.0.1", "port": 3000}, #TODO:Test this
        # 'kafka': {"bootstrap_servers": "localhost:9092", "topic": "knowledge"}, #TODO:Test this
        # 'sqlite': {}  # No configuration required for in-memory SQLite.
    }

    # Benchmark each backend.
    for backend, config in backend_configs.items():
        benchmark_operations(backend, config, num_operations=1000)



# **********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************
from rpio.transformations.transformations import swc2code_py, message2code_py, swc2launch, swc2main, swc2dockerCompose, update_robosapiensIO_ini, add_backbone_config, robochart2aadlmessages, robochart2logical
from rpio.utils.auxiliary import *
from rpio.parsers.parsers import *
from rpio.metamodels.aadl2_IL import *

import configparser
import sys
import os


# ------------------------------------------------------------------------------------
# ------------------------------ CONSTANTS ------------------------------------------
# ------------------------------------------------------------------------------------

BASE_DIR = os.getcwd()

DESIGN_DIR = os.path.join(BASE_DIR, "Design", "design.json")
RPIO_INI_DIR = os.path.join(BASE_DIR,"robosapiensIO.ini")

MESSAGES_DIR = os.path.join(BASE_DIR, "Realization", "Messages")
NODES_DIR = os.path.join(BASE_DIR,  "Realization", "Nodes")
PLATFORM_DIR = os.path.join(BASE_DIR,  "Realization","Platform")
RESOURCES_DIR = os.path.join(BASE_DIR, "Resources")

CONCEPT_DIR = os.path.join(BASE_DIR, "Concept")
MAPLE_RCT = os.path.join(CONCEPT_DIR, "MAPLE-K.rct")
MONITOR_RCT = os.path.join(CONCEPT_DIR, "Monitor.rct")
ANALYSIS_RCT = os.path.join(CONCEPT_DIR, "Analysis.rct")
PLAN_RCT = os.path.join(CONCEPT_DIR, "Plan.rct")
LEGITIMATE_RCT = os.path.join(CONCEPT_DIR, "Legitimate.rct")
EXECUTE_RCT = os.path.join(CONCEPT_DIR, "Execute.rct")
KNOWLEDGE_RCT = os.path.join(CONCEPT_DIR, "Knowledge.rct")

# ------------------------------------------------------------------------------------
# ------------------------------AADL2CODE TASKS --------------------------------------
# ------------------------------------------------------------------------------------
def t_load_design():
    # load name and description from ini
    config = configparser.ConfigParser()
    try:
        config.read(RPIO_INI_DIR)
        name = config['RoboSAPIENSIO']['name']
        description = config['RoboSAPIENSIO']['description']
        try:
            design = system(name=name, description=description, JSONDescriptor=DESIGN_DIR)  # TODO: load AADL when AADL parser is complete
        except:
            print("Design file not found. Please check the path.")
            design = None
    except:
        print("Could not load name and description from ini file")
        design = None
    return design

def t_generate_messages():
    try:
        design = t_load_design()
        # generate messages using the constants for managing systems
        message2code_py(system=design, path=MESSAGES_DIR)
        return True
    except:
        print("Failed to generate the messages")
        return False

def t_generate_swc_skeletons():
    try:
        design = t_load_design()
        # generate swc code skeletons using the constant for nodes directory
        swc2code_py(system=design, path=NODES_DIR)
        return True
    except:
        print("Failed to generate the software components")
        return False

def t_generate_swc_launch():
    try:
        design = t_load_design()
        # generate launch files using constants for platform directories
        swc2launch(system=design.systems[0], path=PLATFORM_DIR)
        return True
    except:
        print("Failed to generate the software component launch files")
        return False

def t_generate_main():
    try:
        config = configparser.ConfigParser()
        # use the constant for the ini file
        config.read(RPIO_INI_DIR)
        packageName = config['PACKAGE']['name']
        prefix = config['PACKAGE']['prefix']
        design = t_load_design()
        # generate main launch file using RESOURCES_DIR instead of a literal "../Resources"
        swc2main(system=design.systems[0], package=packageName, prefix=(prefix if prefix != "" else None), path=RESOURCES_DIR)
        return True
    except:
        print("Failed to generate the software component main file for the given platforms")
        return False

def t_generate_docker():
    try:
        design = t_load_design()
        # generate docker compose file using constant for managing platform directory
        swc2dockerCompose(system=design.systems[0], path=PLATFORM_DIR)
        # add backbone config using RESOURCES_DIR
        add_backbone_config(system=design, path=RESOURCES_DIR)
        return True
    except:
        print("Failed to generate the docker compose for the given platforms")
        return False

def t_update_robosapiensIO_ini():
    try:
        config = configparser.ConfigParser()
        config.read(RPIO_INI_DIR)
        packageName = config['PACKAGE']['name']
        prefix = config['PACKAGE']['prefix']
        design = t_load_design()
        # update ini file using the directory of RPIO_INI_DIR instead of a literal "../"
        update_robosapiensIO_ini(system=design, package=packageName, prefix=prefix, path=os.path.dirname(RPIO_INI_DIR))
        return True
    except:
        print("Could not update robosapiensIO.ini")
        return False

# ------------------------------------------------------------------------------------
# --------------------------RoboChart2AADL TASKS -------------------------------------
# ------------------------------------------------------------------------------------

def t_robochart_to_messages():
    try:
        # Parse RoboChart models using the defined constants
        parser = robochart_parser(
            MAPLEK=MAPLE_RCT,
            Monitor=MONITOR_RCT,
            Analysis=ANALYSIS_RCT,
            Plan=PLAN_RCT,
            Legitimate=LEGITIMATE_RCT,
            Execute=EXECUTE_RCT,
            Knowledge=KNOWLEDGE_RCT
        )
        # generate messages, here DESIGN_DIR is used if it represents the design folder;
        # alternatively
        robochart2aadlmessages(maplek=parser.maplek_model, path=DESIGN_DIR)
        return True
    except:
        print("Failed to generate AADL messages from provided RoboChart models")
        return False

def t_robochart_to_logical():
    try:
        # Parse robochart models
        models_parsed = robochart_parser(MAPLEK='../Concept/MAPLE-K.rct',Monitor='../Concept/Monitor.rct',Analysis='../Concept/Analysis.rct',Plan='../Concept/Plan.rct',Legitimate='../Concept/Legitimate.rct',Execute='../Concept/Execute.rct',Knowledge='../Concept/Knowledge.rct')
        # generate logical architecture
        robochart2logical(parsed=models_parsed,path='../Design')
        print("RoboChart to AADL logical architecture is not implemented yet!")
        return True
    except:
        print("Failed to generate AADL logical architecture from provided RoboChart models")
        return False

# ------------------------------------------------------------------------------------
# ------------------------------ CHECKING TASKS --------------------------------------
# ------------------------------------------------------------------------------------
def t_check_robosapiensio():
    check = check_package_installation(package='robosapiensio')
    return check



import yaml
import logging
import json
import uuid
import pickle
from datetime import datetime
from rpio.clientLibraries.rpclpy.CommunicationManager import CommunicationManager
from rpio.clientLibraries.rpclpy.KnowledgeManager import KnowledgeManager
from rpio.clientLibraries.rpclpy.LoggingAndTracking import LoggingAndTrackingHandler


# ---------------------------------------------------------------------------
# Configuration Keys
# ---------------------------------------------------------------------------
KEY_EVENT_MANAGER_CONFIG = "Event_Manager_Config"
KEY_MESSAGE_MANAGER_CONFIG = "Message_Manager_Config"
KEY_LOGGER_CONFIG = "Logger_Config"
KEY_KNOWLEDGE_CONFIG = "Knowledge_Config"
KEY_LOG_LEVEL = "log_level"
KEY_LOG_FORMAT = "format"
KEY_PROTOCOL = "protocol"

# ---------------------------------------------------------------------------
# Default Values and Messages
# ---------------------------------------------------------------------------
DEFAULT_EVENT_MESSAGE = {}
ENCODING_UTF8 = "utf-8"

WARN_EVENT_MANAGER_NOT_SET = "Event manager is not set for Event publishing."
WARN_REGISTER_EVENT_CALLBACK_NOT_SET = "Event manager is not set for registering event callbacks."

MSG_INITIALIZING_KNOWLEDGE = "Initializing Knowledge: {} knowledge"
MSG_INITIALIZING_EVENT_MANAGER = "Initializing Event Manager"
MSG_INITIALIZING_IO_MANAGER = "Initializing IO Manager"
MSG_USING_COMMUNICATION_MANAGER = "{} is using {} Communication Manager"
MSG_STARTING = "{} is starting..."
MSG_SHUTTING_DOWN = "{} is shutting down..."
MSG_REGISTER_EVENT_CALLBACK = "Registered callback for event: {}"


class Node:
    def __init__(self, config, verbose=False):
        # self.config = self.load_config(config)
        self.config = config
        self.logger = self._initialize_logger()
        self.knowledge = self._initialize_knowledge()  # Initialize knowledge within the component
        self.event_manager = self._initialize_event_manager()
        self.message_manager = self._initialize_message_manager()

        # Initialize MQTT and ROS2 Event
        if self.event_manager:
            self.logger.info(
                MSG_USING_COMMUNICATION_MANAGER.format(
                    self.__class__.__name__,
                    self.config[KEY_EVENT_MANAGER_CONFIG][KEY_PROTOCOL]
                )
            )
        if self.message_manager:
            self.logger.info(
                MSG_USING_COMMUNICATION_MANAGER.format(
                    self.__class__.__name__,
                    self.config[KEY_MESSAGE_MANAGER_CONFIG][KEY_PROTOCOL]
                )
            )

    def load_config(self, config_file):
        with open(config_file, 'r') as file:
            return yaml.safe_load(file)

    def _initialize_logger(self):
        # Create the main logger.
        logger = logging.getLogger(self.__class__.__name__)
        logger.setLevel(self.config[KEY_LOGGER_CONFIG][KEY_LOG_LEVEL])
        # Instantiate the custom logging handler.
        custom_handler = LoggingAndTrackingHandler(self.config[KEY_LOGGER_CONFIG])
        formatter = logging.Formatter(self.config[KEY_LOGGER_CONFIG][KEY_LOG_FORMAT])
        custom_handler.setFormatter(formatter)
        logger.addHandler(custom_handler)
        return logger

    def _initialize_knowledge(self):
        """Initialize the Knowledge object based on the config."""
        self.logger.info(
            MSG_INITIALIZING_KNOWLEDGE.format(
                self.config[KEY_KNOWLEDGE_CONFIG]["knowledge_type"]
            )
        )
        return KnowledgeManager(self.config[KEY_KNOWLEDGE_CONFIG])

    def _initialize_event_manager(self):
        """Initialize the Event Manager based on the config."""
        self.logger.info(MSG_INITIALIZING_EVENT_MANAGER)
        return CommunicationManager(config=self.config[KEY_EVENT_MANAGER_CONFIG])

    def _initialize_message_manager(self):
        """Initialize the IO Manager based on the config."""
        self.logger.info(MSG_INITIALIZING_IO_MANAGER)
        return CommunicationManager(config=self.config[KEY_MESSAGE_MANAGER_CONFIG])

    def start(self):
        """Start the component and enable Event."""
        self.logger.info(MSG_STARTING.format(self.__class__.__name__))
        if self.event_manager:
            self.event_manager.start()
        if self.message_manager:
            self.message_manager.start()

    def shutdown(self):
        """Shutdown the component and stop Event."""
        self.logger.info(MSG_SHUTTING_DOWN.format(self.__class__.__name__))
        if self.event_manager:
            self.event_manager.stop()
        if self.message_manager:
            self.message_manager.stop()
            

    def publish_event(self, event_key, message=DEFAULT_EVENT_MESSAGE):
        """Publish Event using the Event manager."""
        if isinstance(event_key, str):
            # Generate a unique ID for the message
            message_id = str(uuid.uuid4().hex[:8])
            # Get the current timestamp in ISO 8601 format
            timestamp = datetime.now().isoformat()
            
            message['uid'] = message_id
            message['timestamp'] = timestamp
        
            if self.event_manager:
                self.event_manager.publish(event_key,  json.dumps(message))
            else:
                self.logger.warning(WARN_EVENT_MANAGER_NOT_SET)
        else: 
            event_cls = event_key()
            event_cls.uid = str(uuid.uuid4().hex[:8])
            event_cls.timestamp = datetime.now().isoformat()
            value = json.dumps(event_cls.__dict__)
            self.event_manager.publish(event_cls.topic, value)



    def publish_message(self, event_key, message=DEFAULT_EVENT_MESSAGE):
        """Publish Event using the Event manager."""
        if self.message_manager:
            self.message_manager.publish(event_key, message)
        else:
            self.logger.warning(WARN_EVENT_MANAGER_NOT_SET)

    def register_event_callback(self, event_key, callback):
        """Register a callback for Event manager events (MQTT or Redis)."""
        if self.event_manager:
            self.event_manager.subscribe(event_key, callback)
            self.logger.info(MSG_REGISTER_EVENT_CALLBACK.format(event_key))
        else:
            self.logger.warning(WARN_REGISTER_EVENT_CALLBACK_NOT_SET)

    def register_message_callback(self, event_key, callback):
        """Register a callback for Event manager events (MQTT or Redis)."""
        if self.message_manager:
            self.message_manager.subscribe(event_key, callback)
            self.logger.info(MSG_REGISTER_EVENT_CALLBACK.format(event_key))
        else:
            self.logger.warning(WARN_REGISTER_EVENT_CALLBACK_NOT_SET)

    def read_knowledge(self, key, queueSize=1):
        """Read a value from the Knowledge Manager."""
        if isinstance(key, str):
            value = self.knowledge.read(key, queueSize)
            if value is not None:
                value = value.decode(ENCODING_UTF8)  # Convert bytes to string
                try:
                    value = json.loads(value)  # Try to deserialize the value if it's a JSON string
                except json.JSONDecodeError:
                    pass  # If it's not JSON, return it as a string
            
            return value
        else:
            return self.read_knowledge_as_object(key)
    
    def read_knowledge_as_object(self, cls):
        _cls = cls()
        value = self.knowledge.read(_cls.name)
        # if isinstance(value, str):
        #     value = json.loads(value)
        #     # Extract matching arguments based on class __init__
        # try:
        #     obj = cls(**value)
        # except TypeError:
        #     # Fallback: create empty object and manually set attributes
        #     obj = cls.__new__(cls)

        #     for key, value in value.items():
        #         setattr(obj, key, value)
        #     if hasattr(obj, '__init__'):
        #         try:
        #             obj.__init__()
        #         except TypeError:
        #             pass  # if __init__ expects parameters we skip calling it
        # return obj
        return pickle.loads(value)

    def write_knowledge(self, key, value = None):
        """Write a value to the Knowledge Manager."""
        if isinstance(key, str):
            if isinstance(value, str):
                value = str(value)
            return self.knowledge.write(key, value)
        else:
            return self.write_knowledge_as_object(key)
        
    def write_knowledge_as_object(self, cls):
        # """Write a class instance to the Knowledge Manager."""
        # value = json.dumps(cls.__dict__)
        cls.uid = str(uuid.uuid4().hex[:8])
        cls.timestamp = datetime.now().isoformat()
        value = pickle.dumps(cls)
        return self.knowledge.write(cls.name, value)




import threading
import time
import socket

# Constants
KEY_PROTOCOL = "protocol"
PROTOCOL_MQTT = "mqtt"
PROTOCOL_RABBITMQ = "rabbitmq"
PROTOCOL_KAFKA = "kafka"
PROTOCOL_REDIS = "redis"
PROTOCOL_ZENOH = "zenoh"
PROTOCOL_WEBSOCKET = "websocket"
PROTOCOL_TCP = "tcp"
PROTOCOL_TCP_ALT = "tcp/ip"

KEY_BROKER = "broker"
KEY_PORT = "port"
KEY_HOST = "host"
KEY_BOOTSTRAP_SERVERS = "bootstrap_servers"
KEY_URL = "url"

TEST_TOPIC = "test/topic"
DEFAULT_PUBLISH_MESSAGE = "Hello, world!"

WARNING_UNSUPPORTED_PROTOCOL = "Unsupported protocol: {}"
ZENOH_IMPORT_ERROR = "Zenoh library is not installed"
SOCKET_CLOSED_MESSAGE = "Redis listener thread exiting: {}"
WS_NOT_ESTABLISHED = "WebSocket connection not established"
WS_ERROR_PREFIX = "WebSocket error:"
WS_CLOSED_PREFIX = "WebSocket closed"
TCP_RECEIVE_ERROR = "TCP receiving error: {}"

class CommunicationManager:
    def __init__(self, config):
        """
        Initialize the communication manager with the given protocol and configuration.
        Protocol-specific modules are imported here.
        """
        self.protocol = config.get(KEY_PROTOCOL)
        self.config = config
        self.publish_topics = []
        self.subscribe_topics = []
        self.event_callbacks = {}  # Mapping: topic -> callback function
        self._consumer_thread = None

        if self.protocol == PROTOCOL_MQTT:
            import paho.mqtt.client as mqtt
            self.mqtt = mqtt
            self._mqtt_client = None
        elif self.protocol == PROTOCOL_RABBITMQ:
            import pika
            self.pika = pika
            self._rabbitmq_connection = None
            self._rabbitmq_channel = None
        elif self.protocol == PROTOCOL_KAFKA:
            from kafka import KafkaProducer, KafkaConsumer
            self.KafkaProducer = KafkaProducer
            self.KafkaConsumer = KafkaConsumer
            self._kafka_producer = None
            self._kafka_consumer = None
        elif self.protocol == PROTOCOL_REDIS:
            import redis
            self.redis = redis
            self._redis_client = None
            self._redis_pubsub = None
        elif self.protocol == PROTOCOL_ZENOH:
            try:
                import zenoh
            except ImportError:
                zenoh = None
            self.zenoh = zenoh
            self._zenoh_session = None
        elif self.protocol == PROTOCOL_WEBSOCKET:
            import websocket
            self.websocket = websocket
            self._websocket_app = None
            self._websocket_thread = None
        elif self.protocol in [PROTOCOL_TCP, PROTOCOL_TCP_ALT]:
            # Using already imported socket module.
            self._tcp_socket = None
            self._tcp_thread = None
        else:
            raise ValueError(WARNING_UNSUPPORTED_PROTOCOL.format(self.protocol))

    def start(self):
        """Start the client for the selected protocol."""
        if self.protocol == PROTOCOL_MQTT:
            self._start_mqtt()
        elif self.protocol == PROTOCOL_RABBITMQ:
            self._start_rabbitmq()
        elif self.protocol == PROTOCOL_KAFKA:
            self._start_kafka()
        elif self.protocol == PROTOCOL_REDIS:
            self._start_redis()
        elif self.protocol == PROTOCOL_ZENOH:
            self._start_zenoh()
        elif self.protocol == PROTOCOL_WEBSOCKET:
            self._start_websocket()
        elif self.protocol in [PROTOCOL_TCP, PROTOCOL_TCP_ALT]:
            self._start_tcp()
        else:
            raise ValueError(WARNING_UNSUPPORTED_PROTOCOL.format(self.protocol))

    def _start_mqtt(self):
        self._mqtt_client = self.mqtt.Client()
        def on_connect(client, userdata, flags, rc):
            for topic in self.subscribe_topics:
                client.subscribe(topic)
        def on_message(client, userdata, msg):
            if msg.topic in self.event_callbacks:
                self.event_callbacks[msg.topic](msg.payload.decode())
        self._mqtt_client.on_connect = on_connect
        self._mqtt_client.on_message = on_message
        broker = self.config[KEY_BROKER]
        port = self.config[KEY_PORT]
        self._mqtt_client.connect(broker, port, 60)
        self._mqtt_client.loop_start()

    def _start_rabbitmq(self):
        parameters = self.pika.ConnectionParameters(
            host=self.config[KEY_HOST],
            port=self.config[KEY_PORT]
        )
        self._rabbitmq_connection = self.pika.BlockingConnection(parameters)
        self._rabbitmq_channel = self._rabbitmq_connection.channel()
        for topic in self.subscribe_topics:
            self._rabbitmq_channel.queue_declare(queue=topic)
            def callback(ch, method, properties, body, topic=topic):
                if topic in self.event_callbacks:
                    self.event_callbacks[topic](body.decode())
            self._rabbitmq_channel.basic_consume(queue=topic, on_message_callback=callback, auto_ack=True)
        def consume():
            self._rabbitmq_channel.start_consuming()
        self._consumer_thread = threading.Thread(target=consume, daemon=True)
        self._consumer_thread.start()

    def _start_kafka(self):
        bootstrap_servers = self.config[KEY_BOOTSTRAP_SERVERS]
        self._kafka_producer = self.KafkaProducer(bootstrap_servers=bootstrap_servers)
        if self.subscribe_topics:
            self._kafka_consumer = self.KafkaConsumer(*self.subscribe_topics,
                                                       bootstrap_servers=bootstrap_servers,
                                                       auto_offset_reset='earliest',
                                                       consumer_timeout_ms=1000)
            def consume():
                for msg in self._kafka_consumer:
                    if msg.topic in self.event_callbacks:
                        self.event_callbacks[msg.topic](msg.value.decode())
            self._consumer_thread = threading.Thread(target=consume, daemon=True)
            self._consumer_thread.start()

    def _start_redis(self):
        host = self.config[KEY_HOST]
        port = self.config[KEY_PORT]
        self._redis_client = self.redis.Redis(host=host, port=port, decode_responses=True)
        self._redis_pubsub = self._redis_client.pubsub()
        if self.subscribe_topics:
            self._redis_pubsub.subscribe(*self.subscribe_topics)
            def listen():
                try:
                    for message in self._redis_pubsub.listen():
                        if message['type'] == 'message':
                            topic = message['channel']
                            if topic in self.event_callbacks:
                                self.event_callbacks[topic](message['data'])
                except (ValueError, OSError) as e:
                    # Socket was likely closed during shutdown; exit gracefully.
                    print(SOCKET_CLOSED_MESSAGE.format(e))
            self._consumer_thread = threading.Thread(target=listen, daemon=True)
            self._consumer_thread.start()

    def _start_zenoh(self):
        if self.zenoh is None:
            raise ImportError(ZENOH_IMPORT_ERROR)
        config = self.zenoh.Config()
        self._zenoh_session = self.zenoh.open(config)
        for topic in self.subscribe_topics:
            def zenoh_callback(sample, topic=topic):
                payload = sample.payload.decode() if isinstance(sample.payload, bytes) else sample.payload
                if topic in self.event_callbacks:
                    self.event_callbacks[topic](payload)
            self._zenoh_session.declare_subscriber(topic, zenoh_callback)

    def _start_websocket(self):
        ws_url = self.config[KEY_URL]
        def on_message(ws, message):
            if ":" in message:
                parts = message.split(":", 1)
                topic = parts[0].strip()
                payload = parts[1].strip()
                if topic in self.event_callbacks:
                    self.event_callbacks[topic](payload)
            else:
                for cb in self.event_callbacks.values():
                    cb(message)
        def on_error(ws, error):
            print(WS_ERROR_PREFIX, error)
        def on_close(ws, close_status_code, close_msg):
            print(WS_CLOSED_PREFIX, close_status_code, close_msg)
        def on_open(ws):
            print("WebSocket connection opened")
        self._websocket_app = self.websocket.WebSocketApp(ws_url,
                                                            on_message=on_message,
                                                            on_error=on_error,
                                                            on_close=on_close,
                                                            on_open=on_open)
        def run_ws():
            self._websocket_app.run_forever()
        self._websocket_thread = threading.Thread(target=run_ws, daemon=True)
        self._websocket_thread.start()

    def _start_tcp(self):
        host = self.config[KEY_HOST]
        port = self.config[KEY_PORT]
        self._tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._tcp_socket.connect((host, port))
        def listen_tcp():
            while True:
                try:
                    data = self._tcp_socket.recv(1024)
                    if not data:
                        break
                    message = data.decode().strip()
                    if ":" in message:
                        parts = message.split(":", 1)
                        topic = parts[0].strip()
                        payload = parts[1].strip()
                        if topic in self.event_callbacks:
                            self.event_callbacks[topic](payload)
                    else:
                        for cb in self.event_callbacks.values():
                            cb(message)
                except Exception as e:
                    print(TCP_RECEIVE_ERROR.format(e))
                    break
        self._tcp_thread = threading.Thread(target=listen_tcp, daemon=True)
        self._tcp_thread.start()

    def subscribe(self, topic, callback):
        """Register a callback for the given topic."""
        if topic not in self.subscribe_topics:
            self.subscribe_topics.append(topic)
        self.event_callbacks[topic] = callback

        if self.protocol == PROTOCOL_MQTT and self._mqtt_client:
            self._mqtt_client.subscribe(topic)
        elif self.protocol == PROTOCOL_RABBITMQ and self._rabbitmq_channel:
            self._rabbitmq_channel.queue_declare(queue=topic)
            def callback_wrapper(ch, method, properties, body, topic=topic):
                self.event_callbacks[topic](body.decode())
            self._rabbitmq_channel.basic_consume(queue=topic, on_message_callback=callback_wrapper, auto_ack=True)
        elif self.protocol == PROTOCOL_REDIS and self._redis_pubsub:
            self._redis_pubsub.subscribe(topic)
        elif self.protocol == PROTOCOL_ZENOH and self._zenoh_session:
            def zenoh_callback(sample, topic=topic):
                payload = sample.payload.decode() if isinstance(sample.payload, bytes) else sample.payload
                self.event_callbacks[topic](payload)
            self._zenoh_session.declare_subscriber(topic, zenoh_callback)
        # For Kafka, WebSocket, and TCP dynamic subscription is not handled

    def publish(self, topic, message=DEFAULT_PUBLISH_MESSAGE):
        """Publish a message to the specified topic."""
        if topic not in self.publish_topics:
            self.publish_topics.append(topic)
        if self.protocol == PROTOCOL_MQTT and self._mqtt_client:
            result = self._mqtt_client.publish(topic, message)
            return result
        elif self.protocol == PROTOCOL_RABBITMQ and self._rabbitmq_channel:
            self._rabbitmq_channel.queue_declare(queue=topic)
            self._rabbitmq_channel.basic_publish(exchange='', routing_key=topic, body=message)
            return f"Published to {topic}: {message}"
        elif self.protocol == PROTOCOL_KAFKA and self._kafka_producer:
            self._kafka_producer.send(topic, value=message.encode())
            self._kafka_producer.flush()
            return f"Published to {topic}: {message}"
        elif self.protocol == PROTOCOL_REDIS and self._redis_client:
            self._redis_client.publish(topic, message)
            return f"Published to {topic}: {message}"
        elif self.protocol == PROTOCOL_ZENOH and self._zenoh_session:
            self._zenoh_session.put(topic, message.encode())
            return f"Published to {topic}: {message}"
        elif self.protocol == PROTOCOL_WEBSOCKET and self._websocket_app:
            if self._websocket_app.sock and self._websocket_app.sock.connected:
                send_message = f"{topic}: {message}"
                self._websocket_app.send(send_message)
                return f"Published to {topic}: {message}"
            else:
                raise RuntimeError(WS_NOT_ESTABLISHED)
        elif self.protocol in [PROTOCOL_TCP, PROTOCOL_TCP_ALT] and self._tcp_socket:
            send_message = f"{topic}: {message}\n"
            self._tcp_socket.sendall(send_message.encode())
            return f"Published to {topic}: {message}"
        else:
            raise RuntimeError("Client not started or unsupported protocol for publishing.")

    def stop(self):
        """Stop the client and close any open connections."""
        if self.protocol == PROTOCOL_MQTT and self._mqtt_client:
            self._mqtt_client.loop_stop()
            self._mqtt_client.disconnect()
        elif self.protocol == PROTOCOL_RABBITMQ and self._rabbitmq_connection:
            self._rabbitmq_connection.close()
        elif self.protocol == PROTOCOL_KAFKA:
            if self._kafka_consumer:
                self._kafka_consumer.close()
        elif self.protocol == PROTOCOL_REDIS and self._redis_pubsub:
            self._redis_pubsub.close()
        elif self.protocol == PROTOCOL_ZENOH and self._zenoh_session:
            self._zenoh_session.close()
        elif self.protocol == PROTOCOL_WEBSOCKET and self._websocket_app:
            self._websocket_app.close()
            if self._websocket_thread:
                self._websocket_thread.join(timeout=1)
        elif self.protocol in [PROTOCOL_TCP, PROTOCOL_TCP_ALT] and self._tcp_socket:
            self._tcp_socket.close()
        if self._consumer_thread:
            self._consumer_thread.join(timeout=1)


# ----------------- Main Test Script -----------------

def start_tcp_echo_server(host, port):
    """Start a simple TCP echo server."""
    import socket
    def handle_client(client_socket):
        while True:
            data = client_socket.recv(1024)
            if not data:
                break
            client_socket.sendall(data)
        client_socket.close()
    def server_loop():
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((host, port))
        server.listen(5)
        print(f"TCP Echo Server started on {host}:{port}")
        while True:
            client_socket, addr = server.accept()
            client_thread = threading.Thread(target=handle_client, args=(client_socket,), daemon=True)
            client_thread.start()
    server_thread = threading.Thread(target=server_loop, daemon=True)
    server_thread.start()
    return server_thread

def start_ws_echo_server(host, port):
    import asyncio
    import websockets

    async def ws_echo(websocket, path):
        async for message in websocket:
            await websocket.send(message)

    async def server_main():
        async with websockets.serve(ws_echo, host, port):
            await asyncio.Future()  # Run forever

    def run_server():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(server_main())

    ws_thread = threading.Thread(target=run_server, daemon=True)
    ws_thread.start()
    return ws_thread


def test_communication_manager(protocol, config):
    """Test a given protocol by subscribing and publishing on TEST_TOPIC."""
    print(f"\n--- Testing {protocol.upper()} ---")
    received_messages = []
    def callback(msg):
        print(f"[{protocol.upper()}] Received: {msg}")
        received_messages.append(msg)
    cm = CommunicationManager({KEY_PROTOCOL: protocol, **config})
    cm.subscribe(TEST_TOPIC, callback)
    cm.start()
    time.sleep(2)
    result = cm.publish(TEST_TOPIC, DEFAULT_PUBLISH_MESSAGE)
    print(result)
    time.sleep(2)
    cm.stop()
    return received_messages

if __name__ == "__main__":
    # Configuration for each protocol is defined in the main script.
    protocols_config = {
        # PROTOCOL_MQTT: {KEY_BROKER: "localhost", KEY_PORT: 1883},
        # PROTOCOL_RABBITMQ: {KEY_HOST: "localhost", KEY_PORT: 5672},
        # PROTOCOL_KAFKA: {KEY_BOOTSTRAP_SERVERS: "localhost:9092"}, #TODO: Add Kafka configuration, test the Kafka protocol
        PROTOCOL_REDIS: {KEY_HOST: "localhost", KEY_PORT: 6379},
        # PROTOCOL_ZENOH: {},  # Add Zenoh configuration if necessary TODO: Add Zenoh configuration, test the Zenoh protocol
        # PROTOCOL_WEBSOCKET: {KEY_URL: "ws://localhost:8765"},
        # PROTOCOL_TCP: {KEY_HOST: "localhost", KEY_PORT: 9000}
    }

    # # Start WebSocket and TCP echo servers for testing.
    # try:
    #     ws_server = start_ws_echo_server("localhost", 8765)
    #     time.sleep(1)
    # except Exception as e:
    #     print("WebSocket server failed to start:", e)
    #     protocols_config.pop(PROTOCOL_WEBSOCKET, None)

    # try:
    #     tcp_server = start_tcp_echo_server("localhost", 9000)
    #     time.sleep(1)
    # except Exception as e:
    #     print("TCP echo server failed to start:", e)
    #     protocols_config.pop(PROTOCOL_TCP, None)

    results = {}
    for proto, conf in protocols_config.items():
        msgs = test_communication_manager(proto, conf)
        results[proto] = msgs
        print(f"Test for {proto.upper()} received messages: {msgs}")

    print("\nFinal Results:", results)
